from typing import Any

from coastal_dynamics.multiple_choice import MultipleChoiceQuestion
from coastal_dynamics.multiple_selection import MultipleSelectionQuestion
from coastal_dynamics.numeric import NumericQuestion
from coastal_dynamics.text import TextQuestion


class QuestionFactory:
    """
    A factory class for creating question widgets based on question data.

    Attributes:
        question_data (Dict[str, Any]): Dictionary containing data for the question.

    Methods:
        create_question_widget(): Creates a question widget based on the question data.
        serve(): Returns the question widget
    """

    def __init__(self, question_key, question_data: dict[str, Any], serve=True):
        self.question_key = question_key
        self.question_data = question_data
        self.question_widget = self.create_question_widget()

        if serve:
            self.question_widget.serve(question_key)

    def create_question_widget(self):
        """Method that calls the correct creater function based on the question type."""

        question_type = self.question_data.get("type")
        valid_types = {
            "multiple_choice": self._create_multiple_choice_question,
            "multiple_selection": self._create_multiple_selection_question,
            "numeric": self._create_numeric_question,
            "text": self._create_text_question,
        }

        create_func = valid_types.get(question_type)
        if not create_func:
            msg = f"Unknown question type: {question_type}"
            raise ValueError(msg)

        return create_func()

    def _create_multiple_choice_question(self):
        """
        Method that creates mutliple choice question widget from question data.

        Returns:
            MultipleChoiceQuestion (class)
        """

        required_fields = ["name", "question", "options", "answer", "feedback"]
        self._validate_required_fields(required_fields)

        return MultipleChoiceQuestion(
            question_name=self.question_data["name"],
            question_text=self.question_data["question"],
            question_options=self.question_data["options"],
            question_answer=self.question_data["answer"],
            question_feedback=self.question_data["feedback"],
        )

    def _create_multiple_selection_question(self):
        """
        Method that creates mutliple selection question widget from question data.

        Returns:
            MultipleSelectionQuestion (class)
        """

        required_fields = ["name", "question", "options", "answer", "feedback"]
        self._validate_required_fields(required_fields)

        return MultipleSelectionQuestion(
            question_name=self.question_data["name"],
            question_text=self.question_data["question"],
            question_options=self.question_data["options"],
            question_answers=self.question_data["answer"],
            question_feedback=self.question_data["feedback"],
        )

    def _create_numeric_question(self):
        """
        Method that creates numeric question widget from question data.

        Returns:
            NumericQuestion (class)
        """

        required_fields = ["name", "question", "answer", "feedback"]
        self._validate_required_fields(required_fields)

        return NumericQuestion(
            question_name=self.question_data["name"],
            question_text=self.question_data["question"],
            question_answer=self.question_data["answer"],
            question_feedback=self.question_data["feedback"],
            sig_figs=self.question_data.get(
                "sig_figs", 0
            ),  # Handle precision as optional
        )

    def _create_text_question(self):
        """
        Method that creates text question widget from question data.

        Returns:
            TextQuestion (class)
        """

        required_fields = ["name", "question", "answer", "feedback"]
        self._validate_required_fields(required_fields)

        return TextQuestion(
            question_name=self.question_data["name"],
            question_text=self.question_data["question"],
            question_answer=self.question_data["answer"],
            question_feedback=self.question_data["feedback"],
        )

    def _validate_required_fields(self, required_fields):
        """Method that checks if all question data is present for creating a question widget."""

        for field in required_fields:
            if field not in self.question_data:
                raise ValueError(f"Missing required field: {field} in question data")

    def serve(self, key):
        """
        Delegates serving of question to subclasses

        Returns:
            question widget (pn.Column)
        """
        return self.question_widget.serve(key)


if __name__ == "__main__":

    def example():
        question_data = {
            "name": "Q1-1",
            "type": "multiple_choice",
            "question": (
                "Which of the following is a common along the coast in the middle"
                " latitudes?"
            ),
            "options": {
                "a": "Salt marshes",
                "b": "Coral reefs",
                "c": "Tropical rainforests",
                "d": "Mangrove swamps",
            },
            "answer": "a",
            "feedback": {
                "correct": "Well done, ...",
                "incorrect": "Unforunately that is not correct. Please consider...",
            },
        }

        factory = QuestionFactory(question_data)
        return factory.create_question_widget()

    widget = example()
    print("Done")



import json
import os
from pathlib import Path
from typing import Literal

import panel as pn
import pandas as pd

import coastal_dynamics as cd


class Question:
    """Base class for different types of questions with encryption functionality.

    This class provides common functionalities for question widgets including encryption of answers.

    Attributes:
        name (str): The name of the question widget.
        question_text (str): The text of the question.
        feedback_widget (pn.widgets.StaticText): The widget to display feedback.
        submit_button (pn.widgets.Button): The button to submit the answer.
    """

    def __init__(
        self,
        question_name: str,
        question_text: str,
        question_feedback: dict[Literal["correct", "incorrect"], str],
    ):
        self.name = question_name
        self.question_text = question_text
        self.feedback = question_feedback
        self.create_widgets()

    def create_widgets(self) -> None:
        """Create and initialize the Panel widgets for the question."""
        self.question_widget = pn.widgets.StaticText(
            name=self.name, value=self.question_text, sizing_mode="stretch_width"
        )
        self.submit_button = pn.widgets.Button(name="Submit", width=75)
        self.feedback_widget = pn.widgets.StaticText(sizing_mode="stretch_width")

    def serve(self) -> pn.Column:
        """Serve the question as a Panel column."""
        msg = "This method should be implemented by subclasses"
        raise NotImplementedError(msg)

    def check_answer(self, event) -> None:
        """Check the submitted answer against the correct answer."""
        msg = "This method should be implemented by subclasses"
        raise NotImplementedError(msg)

    def hash_answer(self, answer, question_type, sig_figs=None):
        """Delegate the hashing of an answer to the coastal_dynamics module."""
        return cd.hash_answer(answer, question_type, sig_figs=sig_figs)

    def save_answer(self, key, answer):
        """Add the answer to the question stats"""
        if not os.path.exists(Path("saved_answers.json")):
            with open(Path("saved_answers.json"), "w") as f:
                json.dump({"sample_key": "sample_answer"}, f)

        with open(Path("saved_answers.json"), "r") as f:
            json_data = json.load(f)

        if type(answer) in [str, int, float, list]:
            json_data[key] = answer

        with open(Path("saved_answers.json"), "w") as f:
            json.dump(json_data, f)

        return



from typing import Literal

import panel as pn

import coastal_dynamics as cd
from coastal_dynamics.question import Question


class TextQuestion(Question):
    """
    A class to create and manage a text answer question widget.

    This class creates a text question using Panel widgets.
    It supports a question text and a text answer.

    Attributes:
        correct_answer (str): The correct text answer.
        answer_input (pn.widgets.TextInput): The widget for inputting the answer.
    """

    def __init__(
        self,
        question_name: str,
        question_text: str,
        question_answer: str,
        question_feedback: dict[Literal["correct", "incorrect"], str],
    ):
        self.correct_answer = question_answer
        self.answer_input: pn.widgets.TextInput

        super().__init__(question_name, question_text, question_feedback)

    def create_widgets(self) -> None:
        """Create and initialize the Panel widgets for the question."""
        super().create_widgets()
        self.answer_input = pn.widgets.TextInput(
            placeholder="Enter your answer here..."
        )
        self.submit_button.on_click(self.check_answer)

    def check_answer(self, event) -> None:
        """Check the submitted answer against the correct answer."""
        user_answer = str(self.answer_input.value).lower()
        hashed_user_answer = self.hash_answer(user_answer, "text")

        if hashed_user_answer == self.correct_answer:
            self.feedback_widget.value = self.feedback["correct"]
        else:
            self.feedback_widget.value = self.feedback["incorrect"]

        super().save_answer(self.key, user_answer)

    def serve(self, key) -> pn.Column:
        """Serve the question as a Panel column."""
        self.key = key
        return pn.Column(
            self.question_widget,
            self.answer_input,
            self.submit_button,
            self.feedback_widget,
            pn.pane.Markdown(),  # emtpy markdown to widen distance between multiple questions
            width_policy="max",
        )


if __name__ == "__main__":
    question_data = {
        "question": "What is the relative importance of S2 vs M2?",
        "answer": "Some answer",
    }

    tq = TextQuestion(
        question_name="Q3: Simple textual question",
        question_text=question_data["question"],
        question_answer=cd.hash_answer(question_data["answer"], "text"),
        question_feedback={
            "correct": "Correct! Indeed, ...",
            "incorrect": "Incorrect. Please consider ...",
        },
    )
    tq.serve().show()
    print("done")



from typing import Literal

import numpy as np
import panel as pn

import coastal_dynamics as cd
from coastal_dynamics.question import Question
from coastal_dynamics.utils import round_sig


class NumericQuestion(Question):
    """
    A class to create and manage a numeric answer question widget.

    This class creates a numeric question using Panel widgets.
    It supports a question text, numeric answer, and sig_figs for the answer.

    Attributes:
        correct_answer (float): The correct numeric answer.
        sig_figs (int): The number of significant figures of the numeric answer.
        answer_input (pn.widgets.FloatInput): The widget for inputting the answer.
    """

    def __init__(
        self,
        question_name: str,
        question_text: str,
        question_answer: str,
        question_feedback: dict[Literal["correct", "incorrect"], str],
        sig_figs: int | None = None,
    ):
        self.correct_answer = question_answer
        self.sig_figs = sig_figs
        self.answer_input: pn.widgets.FloatInput

        super().__init__(question_name, question_text, question_feedback)

    def create_widgets(self) -> None:
        """Create and initialize the Panel widgets for the question."""
        super().create_widgets()
        self.answer_input = pn.widgets.FloatInput(name="Your Answer", width=200)
        self.submit_button.on_click(self.check_answer)

    def check_answer(self, event) -> None:
        """Check the submitted answer against the correct answer."""
        try:
            user_answer = float(self.answer_input.value)

            # if self.sig_figs:
            #     user_answer = np.format_float_positional(
            #         user_answer,
            #         precision=self.sig_figs,
            #         unique=False,
            #         fractional=False,
            #         trim="k",
            #     )

            if self.sig_figs:
                user_answer = round_sig(user_answer, self.sig_figs)

            hashed_user_answer = self.hash_answer(
                user_answer, "numeric", sig_figs=self.sig_figs
            )

            if hashed_user_answer == self.correct_answer:
                self.feedback_widget.value = self.feedback["correct"]

            else:
                self.feedback_widget.value = self.feedback["incorrect"]

            super().save_answer(self.key, user_answer)

        except ValueError:
            self.feedback_widget.value = "Please enter a valid number."

    def serve(self, key) -> pn.Column:
        """Serve the question as a Panel column."""
        self.key = key
        return pn.Column(
            self.question_widget,
            self.answer_input,
            self.submit_button,
            self.feedback_widget,
            pn.pane.Markdown(),  # emtpy markdown to widen distance between multiple questions
            width_policy="max",
        )


if __name__ == "__main__":
    question_data = {
        "question": "What is the relative importance of S2 vs M2?",
        "answer": 35,
        "sig_figs": 2,
    }

    nq = NumericQuestion(
        question_name="Q3: Simple numeric question",
        question_text=question_data["question"],
        question_answer=cd.hash_answer(
            np.round_sig(
                float(question_data["answer"]),
                precision=question_data["sig_figs"],
                unique=False,
                fractional=False,
                trim="k",
            ),
            "numeric",
        ),
        question_feedback={
            "correct": "Correct!...",
            "incorrect": "Incorrect, try again. Please consider that...",
        },
        sig_figs=question_data["sig_figs"],
    )
    nq.serve().show()
    print("done")



from typing import Literal

import panel as pn

import coastal_dynamics as cd
from coastal_dynamics.question import Question


class MultipleSelectionQuestion(Question):
    """A class to create and manage multiple selection questions using Panel widgets.

    This class creates a multiple selection question using Panel widgets.
    It supports question text, multiple options, and multiple correct answers.
    The correct answers are stored in an encrypted format for basic obfuscation.

    Attributes:
        options (Dict[str, str]): A dictionary of option keys and their text.
        correct_answers (List[str]): The encrypted correct answer keys.
        options_widget (pn.widgets.CheckBoxGroup): The widget for displaying options.
        options_inverse (Dict[str, str]): Inverse mapping of options for easy lookup.
    """

    def __init__(
        self,
        question_name: str,
        question_text: str,
        question_options: dict[str, str],
        question_answers: list[str],
        question_feedback: dict[Literal["correct", "incorrect"], str],
    ):
        self.options = question_options
        self.correct_answers = question_answers
        self.options_widget: pn.widgets.CheckBoxGroup
        self.options_inverse = {v: k for k, v in self.options.items()}

        super().__init__(question_name, question_text, question_feedback)

    def create_widgets(self) -> None:
        """Create and initialize the Panel widgets for the question."""
        super().create_widgets()
        self.options_widget = pn.widgets.CheckBoxGroup(
            name="Options", options=list(self.options.values())
        )
        self.submit_button.on_click(self._check_answers)

    def _check_answers(self, event: pn.widgets.Button) -> None:
        """Check the selected answers against the correct ones."""
        selected_options = [
            self.options_inverse[opt] for opt in self.options_widget.value
        ]

        if set(cd.hash_answer(selected_options, "multiple_selection")) == set(
            self.correct_answers
        ):
            self.feedback_widget.value = self.feedback["correct"]
        else:
            self.feedback_widget.value = self.feedback["incorrect"]

        super().save_answer(self.key, self.options_widget.value)

    def serve(self, key) -> pn.Column:
        """Serve the complete question widget."""
        self.key = key
        return pn.Column(
            self.question_widget,
            self.options_widget,
            self.submit_button,
            self.feedback_widget,
            pn.pane.Markdown(),  # emtpy markdown to widen distance between multiple questions
            width_policy="max",
        )


if __name__ == "__main__":
    question_data = {
        "question": "Select all features commonly found along a coastline",
        "options": {
            "a": "Beaches",
            "b": "Glaciers",
            "c": "Estuaries",
            "d": "Mountains",
        },
        "answer": ["a", "c"],  # Multiple correct answers
    }

    msq = MultipleSelectionQuestion(
        question_name="Q1: Coastline Features Quiz",
        question_text=question_data["question"],
        question_options=question_data["options"],
        question_answers=cd.hash_answer(question_data["answer"], "multiple_selection"),
        question_feedback={
            "correct": (
                "Correct! Beaches and estuaries are common features of coastlines."
            ),
            "incorrect": "Incorrect! Try again. Please consider..",
        },
    )
    print("done")



import os
from pathlib import Path
import json

import panel as pn
import pandas as pd

from coastal_dynamics.factory import QuestionFactory


class QuestionIndustry:
    """
    An industry class for creating and managing question widgets for a jupyter notebook.

    Attributes:
        question_data (Dict[str, Any]): Dictionary containing data for all questions in a notebook

    Methods:
        serve: Serves a pn.column concatenating multiple questions.
    """

    def __init__(self, question_dict):
        self.question_dict = question_dict

        self.served_questions = []
        self.question_counter = 1

    def serve(self, key_list):
        """
        Serves a pn.column concatenating multiple questions.

        Arguments:
            key_list: list with keys referring to specific questions in the question database (.json)

        Returns:
            question widget (pn.Column) containing multiple questions
        """
        # initialize column
        question_col = pn.Column(
            width_policy="max"
        )  # note that the width_policy='max' kwarg is added within the serve function of each question type individually

        # loop through questions to be added
        for question_key in key_list:
            if not question_key in self.served_questions:
                self.served_questions.append(question_key)

                self.question_dict[question_key][
                    "name"
                ] = f"Q-{int(self.question_counter)}"

                self.question_counter += 1

            question = QuestionFactory(
                question_key, self.question_dict[question_key]
            ).serve(question_key)

            # check if any previous answer are available
            if not os.path.exists(Path("saved_answers.json")):
                with open(Path("saved_answers.json"), "w") as f:
                    json.dump({"sample_key": "sample_answer"}, f)

            with open(Path("saved_answers.json"), "r") as f:
                json_data = json.load(f)

            if question_key in json_data.keys():
                question[1].value = json_data[question_key]

            question_col.append(question)

        return question_col



"""Tools for interactive teaching and learning with Jupyter notebooks."""

__author__ = (
    """TU Delft Coastal Dynamics, Floris Calkoen, Kevin de Bruijn, Judith Bosboom"""
)
__email__ = "j.bosboom@tudelft.nl"
__version__ = "1.0.2"

from .factory import QuestionFactory
from .io import read_questions, write_questions
from .multiple_choice import MultipleChoiceQuestion
from .multiple_selection import MultipleSelectionQuestion
from .numeric import NumericQuestion
from .question import Question
from .text import TextQuestion
from .utils import hash_answer, launch_app, load_notebook, save_notebook

from .industry import QuestionIndustry
from .modify_notebooks import clear_answers
from .use_answers_app import UseAnswersApp

__all__ = [
    "MultipleChoiceQuestion",
    "MultipleSelectionQuestion",
    "TextQuestion",
    "NumericQuestion",
    "QuestionFactory",
    "Question",
    "hash_answer",
    "launch_app",
    "load_notebook",
    "save_notebook",
    "read_questions",
    "write_questions",
    "QuestionIndustry",
    "clear_answers",
    "UseAnswersApp",
]



import json
import os
from typing import Optional

import adlfs
import fsspec


def read_questions(
    blob_name: str, storage_options: Optional[dict[str, str]] = None
) -> dict:
    """
    Reads a JSON file from a local filesystem or Azure Blob storage.

    Args:
        blob_name (str): The blob name or local file path.
        storage_options (Optional[Dict[str, str]]): If given, contains options such as
            account name and SAS token for Azure Blob storage.

    Returns:
        Dict: The content of the JSON file.
    """
    # Use fsspec to handle both local and Azure Blob storage cases
    with fsspec.open(blob_name, "r", **(storage_options or {})) as f:
        questions = json.load(f)
    return questions


# BUG: will not write to cloud storage, but local stoarge. Keep here as template
# for later work. So this function (using fsspec) will replace write_questions
# below to make it morge storage agnostic.

# def write_questions(
#     processed_questions: dict,
#     blob_name: str,
#     storage_options: Optional[dict[str, str]] = None,
# ) -> None:
#     """
#     Writes a JSON file to a local filesystem or Azure Blob storage.

#     Args:
#         processed_questions (dict): The content to be written to the JSON file.
#         blob_name (str): The blob name or local file path.
#         storage_options (Optional[Dict[str, str]]): If given, contains options such as
#             account name and SAS token for Azure Blob storage.
#     """
#     # Use fsspec to handle both local and Azure Blob storage cases
#     with fsspec.open(blob_name, "w", **(storage_options or {})) as f:
#         json.dump(processed_questions, f, indent=4)


# The function below was used previously to write to cloud storage, but is replaced now
# to write to local storage

# def write_questions(processed_questions, blob_name, storage_options):
#     fs = adlfs.AzureBlobFileSystem(**storage_options)
#     with fs.open(f"{blob_name}", mode="w") as f:
#         json.dump(processed_questions, f, indent=4)


def write_questions(processed_questions, hashed_blob_name, storage_options, storage):
    """
    Writes questions to .json file, either to an online repo (storage="azure") or locally (storage="local").

    UPDATE: storage="azure" has become outdated and is no longer funcitonal.
    """
    if storage == "azure":
        fs = adlfs.AzureBlobFileSystem(**storage_options)

        with fs.open(f"{hashed_blob_name}", mode="w") as f:
            json.dump(processed_questions, f, indent=4)

    elif storage == "local":
        with open(hashed_blob_name, "w") as f:
            json.dump(processed_questions, f, indent=4)

    return



def clear_answers(
    notebook, begin_key="#BEGIN_CORRECT_ANSWER", end_key="#END_CORRECT_ANSWER"
):
    """Modify the notebook (i.e., remove all notebooks for a given beginning and end key).
    A notebook consists of a dictionary with keys:
        - "cells" (containing a list of cells)
        - "metadata"
        - "nbformat"
        - "nbformat_minor"
    We need to look at the "source" key of each cell in the list of cells.
    """

    number_of_answers_removed = 0

    # loop through cells
    for i_cell in range(len(notebook["cells"])):
        source = notebook["cells"][i_cell]["source"]

        for i, begin_line in enumerate(source):
            if begin_key in begin_line:
                for j, end_line in enumerate(source[i:]):
                    if end_key in end_line:
                        source = source[:i] + source[i + j + 1 :]

                        number_of_answers_removed += 1

                        break
                break

        notebook["cells"][i_cell]["source"] = source

    return notebook, number_of_answers_removed



from typing import Any, Dict, List, Union

import geopandas as gpd
import shapely
from pyproj import Transformer


def build_bbox(min_lon, min_lat, max_lon, max_lat, src_crs, crs):
    """
    Build a latitude / longitude bounding box from the coordinates.
    """
    # minimum / maximum points from the dataset
    # left, bottom, right, top = (-5802250.0, -622000.0, -5519250.0, -39000.0)
    points = [
        [min_lon, max_lon, max_lon, min_lon],
        [min_lat, min_lat, max_lat, max_lat],
    ]

    transformer = Transformer.from_crs(src_crs, crs, always_xy=True)
    lons, lats = transformer.transform(*points)

    west = min(lons)
    east = max(lons)
    north = max(lats)
    south = min(lats)
    return [west, south, east, north]


def geometry_to_bbox(geometry: Dict[str, Any]) -> List[float]:
    """Extract the bounding box from a geojson geometry

    Args:
        geometry : GeoJSON geometry dict

    Returns:
        list: Bounding box of geojson geometry, formatted according to:
        https://tools.ietf.org/html/rfc7946#section-5
    """
    coords = geometry["coordinates"]

    lats: List[float] = []
    lons: List[float] = []

    def extract_coords(coords: List[Union[List[float], List[List[Any]]]]) -> None:
        for x in coords:
            # This handles points
            if isinstance(x, float):
                assert isinstance(
                    coords[0], float
                ), f"Type mismatch: {coords[0]} is not a float"
                assert isinstance(
                    coords[1], float
                ), f"Type mismatch: {coords[1]} is not a float"
                lats.append(coords[0])
                lons.append(coords[1])
                return
            if isinstance(x[0], list):
                extract_coords(x)  # type:ignore
            else:
                lat, lon = x
                lats.append(lat)  # type:ignore
                lons.append(lon)  # type:ignore

    extract_coords(coords)

    lons.sort()
    lats.sort()

    bbox = [lats[0], lons[0], lats[-1], lons[-1]]

    return bbox


def bbox_to_geometry(bbox: List[float]) -> Dict:
    """Use shapely.geometry.shape to cast to shapeley geom"""
    return {
        "type": "Polygon",
        "coordinates": [
            [
                [bbox[2], bbox[1]],
                [bbox[2], bbox[3]],
                [bbox[0], bbox[3]],
                [bbox[0], bbox[1]],
                [bbox[2], bbox[1]],
            ]
        ],
    }


def geo_bbox(
    min_lon: float,
    min_lat: float,
    max_lon: float,
    max_lat: float,
    src_crs="EPSG:4326",
    dst_crs="EPSG:4326",
) -> gpd.GeoDataFrame:
    """GeoDataFrame with a bounding box that can be used for various gis operations.
    Args:
        min_lon (float): most eastern longitude
        min_lat (float): most northern latitude
        max_lon (float): most wester longitude
        max_lat (float): most soutern latitude
        src_crs (str, optional): Valid EPSG string or number (int). Defaults to "EPSG:4326".
        dst_crs (str, optional): Valid EPSG string or number (int). Defaults to "EPSG:4326".

    Returns:
        gpd.GeoDataFrame: GeoDataFrame with bounding box as geometry.
    """

    bbox = [min_lon, min_lat, max_lon, max_lat]
    bbox = bbox_to_geometry(bbox)
    bbox = shapely.geometry.shape(bbox)
    return gpd.GeoDataFrame(geometry=[bbox], crs=src_crs).to_crs(dst_crs)


def get_xy_range(gdf):
    """min, max lon/lat formatted as xy range for geoviews."""
    x_range = tuple(gdf.total_bounds[[0, 2]])
    y_range = tuple(gdf.total_bounds[[1, 3]])
    return x_range, y_range



from typing import Literal

import panel as pn

import coastal_dynamics as cd
from coastal_dynamics.question import Question


class MultipleChoiceQuestion(Question):
    """A class to create and manage a multiple choice question widget.

    This class creates a multiple choice question using Panel widgets.
    It supports question text, multiple options, and a single correct answer.
    The correct answer is stored in an encrypted format for basic obfuscation.

    Attributes:
        options (Dict[str, str]): A dictionary of option keys and their text.
        correct_answer (str): The encrypted correct answer key.
        options_widget (pn.widgets.RadioBoxGroup): The widget for displaying the options.
        options_inverse (Dict[str, str]): Inverse mapping of options for easy lookup.
    """

    def __init__(
        self,
        question_name: str,
        question_text: str,
        question_options: dict[str, str],
        question_answer: str,
        question_feedback: dict[Literal["correct", "incorrect"], str],
    ):
        self.options = question_options
        self.correct_answer = question_answer
        self.options_widget: pn.widgets.RadioBoxGroup
        self.options_inverse = {v: k for k, v in self.options.items()}

        super().__init__(question_name, question_text, question_feedback)

    def create_widgets(self) -> None:
        """Create and initialize the Panel widgets for the question."""
        super().create_widgets()
        self.options_widget = pn.widgets.RadioBoxGroup(
            name="Options", options=list(self.options.values())
        )
        self.submit_button.on_click(self._check_answer)

    def _check_answer(self, event: pn.widgets.Button) -> None:
        """Check the selected answer against the correct answer."""
        selected_option = self.options_inverse[self.options_widget.value]
        if (
            self.hash_answer(selected_option, question_type="multiple_choice")
            == self.correct_answer
        ):
            self.feedback_widget.value = self.feedback["correct"]
        else:
            self.feedback_widget.value = self.feedback["incorrect"]
        super().save_answer(self.key, self.options_widget.value)

    def serve(self, key) -> pn.Column:
        """Serve the question as a Panel column."""
        self.key = key
        return pn.Column(
            self.question_widget,
            self.options_widget,
            self.submit_button,
            self.feedback_widget,
            pn.pane.Markdown(),  # emtpy markdown to widen distance between multiple questions
            width_policy="max",
        )


if __name__ == "__main__":
    question_data = {
        "question": (
            "Which coastal system do you typically find in tide-dominated coasts?"
        ),
        "options": {
            "a": "Mudflats",
            "b": "Open coasts",
            "c": "Cliffed coasts",
            "d": "Mixed sand and gravel beaches",
        },
        "answer": cd.hash_answer("a", "multiple_choice"),  # Multiple correct answers
    }

    mcq = MultipleChoiceQuestion(
        question_name="Q1: Coastline Features Quiz",
        question_text=question_data["question"],
        question_options=question_data["options"],
        question_answer=question_data["answer"],
        question_feedback={
            "correct": "Well done, ...",
            "incorrect": "Unforunately that is not correct. Please consider...",
        },
    )
    mcq.serve().show()
    print("done")



from pathlib import Path
import json
import os

import panel as pn


class UseAnswersApp:
    """
    Class for creating and serving an application that allows users to use or 'reset' their previously given answers.
    """

    def __init__(self, notebook_prefix="1a"):
        self.saved_answers_path = Path("./saved_answers.json")

        # get notebook prefix
        self.notebook_prefix = notebook_prefix

    def serve(self):
        """
        Function for building the application.

        Returns:
            application (pn.Column)
        """
        return self.build_app()

    def get_number_saved_answers(self):
        """
        Computes the number of given answers present in any previous saves.

        Returns:
            number of given answers (int)
        """
        # check if the file exists
        if os.path.exists(self.saved_answers_path):
            # read the file
            with open(self.saved_answers_path, "r") as f:
                saved_answers = json.load(f)

            # get the number of questions that have previously been answered
            n_questions = len(
                [
                    key
                    for key in list(saved_answers.keys())
                    if self.notebook_prefix in key
                ]
            )

        else:
            n_questions = 0

        return n_questions

    def build_app(self):
        """
        Build the app (with all text and buttons).

        Returns:
            app (pn.Column)
        """
        self.n_questions = self.get_number_saved_answers()
        self.n_questions_removed = 0

        self.explanation_text = pn.pane.Markdown(
            f"""
            ### Use saved answers

            We found saved answers for {self.n_questions} questions.
            """
        )

        self.help_text = str(
            """
            If you want to use your saved answers, you can skip this interface and move on to the rest of the notebook. This will load your previous answers and use them to pre-fill the questions in this notebook (as soon as you run the corresponding cells). You can then change the answers if you want to.

            If you want to disregard your saved answers, select the button "Disregard saved answers". This will delete your previous answers (for this notebook) and you will have to answer all (non-coding) questions again.
            """,
        )
        self.help_text_widget = pn.widgets.StaticText(value=" ")

        self.help_button = pn.widgets.Button(name="Help", button_type="default")

        # self.use_saved_answers_button = pn.widgets.Button(
        #     name="Use saved answers", button_type="default"
        # )
        self.disregard_saved_answers_button = pn.widgets.Button(
            name="Disregard saved answers", button_type="default"
        )
        self.are_you_sure_button = pn.widgets.Button(
            name="Are you sure?", button_type="danger", visible=False
        )

        self.ready_to_go_text_widget = pn.widgets.StaticText(value=" ")
        self.ready_to_go_text = """You can now continue with the notebook."""

        @pn.depends(
            self.help_button.param.value,
            # self.use_saved_answers_button.param.value,
            self.disregard_saved_answers_button.param.value,
            self.are_you_sure_button.param.value,
        )
        def main(
            help_button_value,
            # use_saved_answers_button_value,
            disregard_saved_answers_button_value,
            are_you_sure_button_value,
        ):
            """Function to keep track of button presses.

            Arguments:
                help_button_value (bool)
                use_saved_answers_button_value (bool) --> OUTDATED
                disregard_saved_answers_button_value (bool)
                are_you_sure_button_value (bool)

            """
            if help_button_value:
                self.help_text_widget.value = self.help_text
            else:
                str(" ")

            # if use_saved_answers_button_value:
            #     self.ready_to_go_text = (
            #         f"{self.n_questions} saved answers were loaded."
            #         + self.ready_to_go_text
            #     )
            #     self.disable_widgets()

            if disregard_saved_answers_button_value:
                self.are_you_sure_button.visible = True

            if are_you_sure_button_value:
                self.remove_question_keys()

            return None

        app = pn.Column(
            self.explanation_text,
            self.help_button,
            self.help_text_widget,
            # self.use_saved_answers_button,
            pn.Row(self.disregard_saved_answers_button, self.are_you_sure_button),
            self.ready_to_go_text_widget,
            main,
        )

        return app

    def disable_widgets(self):
        """
        Function to disable all widgets.

        Returns:
            True
        """
        for widget in [
            # self.use_saved_answers_button,
            self.disregard_saved_answers_button,
            self.are_you_sure_button,
        ]:
            widget.disabled = True

        self.ready_to_go_text_widget.value = self.ready_to_go_text

        return True

    def remove_question_keys(self, fp=Path("saved_answers.json")):
        """
        Function to remove previously saved answers (i.e., question keys) from served question apps.
        """
        if os.path.exists(fp):
            # read the file
            with open(self.saved_answers_path, "r") as f:
                saved_answers = json.load(f)

            # remove the question keys
            for key in list(saved_answers.keys()):
                if self.notebook_prefix in key:
                    saved_answers.pop(key)
                    self.n_questions_removed += 1

            # write the file again
            with open(self.saved_answers_path, "w") as f:
                json.dump(saved_answers, f)

            self.ready_to_go_text = (
                f"{self.n_questions_removed} answers were removed. "
                + self.ready_to_go_text
            )

        self.disable_widgets()

    def rename_saved_answers(self, fname):
        """Function to rename saved answer file to a new file."""
        with open(self.saved_answers_path, "r") as f:
            saved_answers = json.load(f)

        with open(fname, "w") as f:
            json.dump(saved_answers, f)

        self.remove_question_keys()

        self.disable_widgets()



import hashlib
import socket
import os
import logging
import warnings
import json

from math import log10, floor

import numpy as np


def hash_answer(answer, question_type, sig_figs=None):
    """Hash a single answer or a list of answers based on the question type."""
    if question_type == "multiple_selection":
        # For multiple_selection, directly hash each answer in the list
        return [hashlib.sha256(ans.encode()).hexdigest() for ans in answer]
    elif question_type == "text":
        # For text questions, normalize to lower case before hashing
        return hashlib.sha256(answer.lower().encode()).hexdigest()
    elif question_type == "multiple_choice":
        # For multiple_choice and numeric, directly hash the answer
        return hashlib.sha256(str(answer).encode()).hexdigest()
    elif question_type == "numeric":
        # if sig_figs:
        #     answer = np.format_float_positional(
        #         float(answer),
        #         precision=sig_figs,
        #         unique=False,
        #         fractional=False,
        #         trim="k",
        #     )
        if sig_figs:
            answer = round_sig(
                float(answer),
                sig_figs,
            )
        return hashlib.sha256(str(answer).encode()).hexdigest()
    else:
        msg = f"Unsupported question type: {question_type}"
        raise ValueError(msg)


def round_sig(x, sig):
    """Round a given number x to a certain amount of significant figures, and return this number in as float"""
    # Handle zero input explicitly
    if x == 0:
        return 0

    # Calculate the rounding precision
    precision = sig - int(floor(log10(abs(x)))) - 1

    # Compute rounded value
    factor = 10**precision
    adjusted_x = x * factor

    if adjusted_x % 1 == 0.5:  # Check if it ends in an exact 5
        adjusted_x = adjusted_x + 0.5  # Push up if it's exactly halfway

    return round(adjusted_x) / factor


def find_free_port():
    """Finds an available port to serve a panel app."""
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(("", 0))  # Bind to an available port
        return s.getsockname()[1]  # Get assigned port


def launch_app(app):
    """Launches panel app, returns string with address of app"""

    # Turn of warnings in notebooks
    logging.getLogger().setLevel(logging.ERROR)
    warnings.filterwarnings("ignore")

    # find available port
    port = find_free_port()

    # launch app
    app.show(port=port, threaded=True, verbose=False)

    # generate printed address
    if "JUPYTERHUB_USER" not in list(os.environ.keys()):
        address = f"http://localhost:{port}"

    else:
        user = os.environ["JUPYTERHUB_USER"]

        address = f"https://coastal.citg.tudelft.nl/codebook/user/{user}/proxy/{port}/"

    return "See the app at: " + address


def load_notebook(file_path):
    """Load notebook as json"""
    with open(file_path, "r", encoding="utf-8") as f:
        notebook = json.load(f)
    return notebook


def save_notebook(notebook, new_file_path):
    """Save the modified notebook"""
    with open(new_file_path, "w", encoding="utf-8") as f:
        json.dump(notebook, f, indent=4, ensure_ascii=False)
    return


# NOT NEEDED ANYMORE SINCE WE ARE USING RELATIVE FILE PATHS EVERYWHERE
# def path(fpath):
#     """Returns path to the database files, either for global users of the jupyter hub
#     or when using a local file path.

#     fpath (str): string containing local file path
#         example: "database/1_coastal_classification/1_earthquakes_sample.parquet"
#     """


#     # users that run locally
#     if "JUPYTERHUB_USER" not in list(os.environ.keys()):
#         p = "./" + fpath

#     # users of HUB
#     else:
#         p = "/var/" + fpath

#     return Path(p)
